//---------------------------------------------------------------------------

#include <basepch.h>

#pragma hdrstop

#include "D2XXDriver.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace D2xxdriver
{
        void __fastcall PACKAGE Register()
        {
                 TComponentClass classes[1] = {__classid(TD2XXDriver)};
                 RegisterComponents("Klen Tools", classes, 0);
        }
}
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//
//---------------------------------------------------------------------------
static inline void ValidCtrCheck(TD2XXDriver *)
{
   new TD2XXDriver ((TComponent*)NULL) ;
}

//---------------------------------------------------------------------------
__fastcall TD2XXDriver::TD2XXDriver(TComponent* Owner) : TComponent(Owner)
{
  FHandle = INVALID_HANDLE_VALUE;
  FFTDeviceName = "" ;
  FBaudRate = cbr9600;
  FParity = paNone;
  FStopBits = sb1_0;
  FDataBits = 8;
  FComControl = ccNone;
  FMonitorEvents << evBreak << evCTS << evDSR << evError << evRing
                 << evRlsd << evRxChar << evTxEmpty;

  IBuffSize = 2048;
  OBuffSize = 4096;
  IBuffUsed = 0;
  OBuffUsed = 0;

  FRTS = false;
  FDTR = false;

  FXonLim = 0;
  FXoffLim = 1024;

  FXonChar = 0x11;
  FXoffChar = 0x13;

  FEventChar = 0;

  InitializeCriticalSection(&WriteSection);
  InitializeCriticalSection(&ReadSection);
  rtEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
  wtEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
  reEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
  seEvent = CreateEvent(NULL, FALSE, FALSE, NULL);

  Status = 0;

  FDCB.DCBlength = sizeof(FTDCB) ;

}

//---------------------------------------------------------------------------
__fastcall TD2XXDriver::~TD2XXDriver(void)
{
  ClosePort();

  CloseHandle(seEvent);
  CloseHandle(reEvent);
  CloseHandle(wtEvent);
  CloseHandle(rtEvent);
 
  DeleteCriticalSection(&ReadSection);
  DeleteCriticalSection(&WriteSection);
}

//---------------------------------------------------------------------------
bool __fastcall TD2XXDriver::IsEnabled(void)
{
  return( FHandle != INVALID_HANDLE_VALUE );
}

//---------------------------------------------------------------------------
int __fastcall TD2XXDriver::GetOutQueCount(void)
{

  DWORD lpErrors;
  FTCOMSTAT lpStat;

  int Result = 0;
  if( IsEnabled() ){
    FT_W32_ClearCommError( FHandle, &lpErrors, &lpStat );
    Result = lpStat.cbOutQue;
  }
  return(Result);
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::OpenPort(void)
{
  if( FOpen )  return;

  IBuffUsed = 0;
  OBuffUsed = 0;

  OBuffer = new unsigned char[OBuffSize];
  IBuffer = new unsigned char[IBuffSize];
  SetLastError(0);

  FHandle = FT_W32_CreateFile(FFTDeviceName.c_str(),GENERIC_READ|GENERIC_WRITE,0,0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED |FT_OPEN_BY_DESCRIPTION,0);

  if( IsEnabled() ){
    FTTIMEOUTS CommTimeOuts;

    

    FT_W32_SetupComm( FHandle, IBuffSize, OBuffSize );

    ZeroMemory( &CommTimeOuts, sizeof(FTTIMEOUTS) );
    CommTimeOuts.ReadIntervalTimeout = 100;
    CommTimeOuts.ReadTotalTimeoutMultiplier = 1;
    CommTimeOuts.ReadTotalTimeoutConstant = 30;
    CommTimeOuts.WriteTotalTimeoutMultiplier = 10;
    CommTimeOuts.WriteTotalTimeoutConstant = 0;


    FT_W32_SetCommTimeouts( FHandle, &CommTimeOuts );

    SetComControl(FComControl);
    SetBaudRate(FBaudRate);
    SetParity(FParity);
    SetStopBits(FStopBits);
    SetDataBits(FDataBits);
    SetXonLim(FXonLim);
    SetXoffLim(FXoffLim);
    SetXonChar(FXonChar);
    SetXoffChar(FXoffChar);
    SetEventChar(FEventChar);


    StatusThread = new TD2XXStatusThread( this, FMonitorEvents );

    ResetEvent(rtEvent);
    ReadThread = new TD2XXReadThread( this );

    ResetEvent(wtEvent);
    WriteThread = new TD2XXWriteThread( this );

    ResetEvent(reEvent);
    ReadEventThread = new TD2XXReadEventThread( this );

    ResetEvent(seEvent);
    StatusEventThread = new TD2XXStatusEventThread( this );

    FOpen = true;

    SetDTR(FDTR);
    SetRTS(FRTS);

    if( FOnOpen )  FOnOpen( this, GetLastError() );
  }else{
     //MessageBox(NULL ,SysErrorMessage(  GetLastError()).c_str(), "  ",  MB_OK | MB_ICONSTOP );
     throw "  " ;
  }
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::ClosePort(void)
{
  if( !FOpen )   return;

    FCTS  = False;
    FDSR  = False;
    FDCD  = False;
    FRing = False;

    StatusThread->Terminate();
    ResetEvent(StatusThread->SOL.hEvent);
    FT_W32_SetCommMask(FHandle, 0);
           TerminateThread( (void*)StatusThread->Handle , 0 ) ;
           //StatusThread->WaitFor();
    delete StatusThread;
    StatusThread = NULL;

    ReadThread->Terminate();
    ResetEvent(ReadThread->ROL.hEvent);
    SetEvent(rtEvent);
          TerminateThread( (void*)ReadThread->Handle , 0 ) ;
          //ReadThread->WaitFor();
    delete ReadThread;
    ReadThread = NULL;

    WriteThread->Terminate();
    ResetEvent(WriteThread->WOL.hEvent);
    SetEvent(wtEvent);
         //TerminateThread( (void*)WriteThread->Handle , 0 ) ;
         WriteThread->WaitFor();
    delete WriteThread;
    WriteThread = NULL;

    ReadEventThread->Terminate();
    SetEvent(reEvent);
         //TerminateThread( (void*)ReadEventThread->Handle , 0 ) ;
         ReadEventThread->WaitFor();
    delete ReadEventThread;
    ReadEventThread = NULL;

    StatusEventThread->Terminate();
    SetEvent(seEvent);
         //TerminateThread( (void*)StatusEventThread->Handle , 0 ) ;
         StatusEventThread->WaitFor();
    delete StatusEventThread;
    StatusEventThread = NULL;

    FT_W32_CloseHandle(FHandle);
    FHandle = INVALID_HANDLE_VALUE;

    delete IBuffer;
    delete OBuffer;
    IBuffer = NULL;
    OBuffer = NULL;
    IBuffUsed = 0;
    OBuffUsed = 0;

    FOpen = false;

  if( FOnClose )  FOnClose( this );
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetFFTDeviceName( AnsiString FTDeviceName )
{
  Open = false;
  FFTDeviceName = FTDeviceName ;
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetIBuffSize( DWORD _size )
{
  bool tmp = Open;
  Open = false;
  IBuffSize = _size;
  Open = tmp;
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetOBuffSize( DWORD _size )
{
  bool tmp = Open;
  Open = false;
  OBuffSize = _size;
  Open = tmp;
}

//---------------------------------------------------------------------------
DWORD __fastcall TD2XXDriver::GetInBuffUsed(void)
{
  return( IBuffUsed );
}

//---------------------------------------------------------------------------
DWORD __fastcall TD2XXDriver::GetInBuffFree(void)
{
  EnterCriticalSection( &ReadSection );
  int Result = IBuffSize - IBuffUsed;
  LeaveCriticalSection( &ReadSection );
  return( Result );
}

//---------------------------------------------------------------------------
DWORD __fastcall TD2XXDriver::GetOutBuffUsed(void)
{
  return( OBuffUsed );
}

//---------------------------------------------------------------------------
DWORD __fastcall TD2XXDriver::GetOutBuffFree(void)
{
  EnterCriticalSection( &WriteSection );
  int Result = OBuffSize - OBuffUsed;
  LeaveCriticalSection( &WriteSection );
  return( Result );
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetOpen( bool o )
{
  if( FOpen == o )  return;

  if( o )
    OpenPort();
  else
    ClosePort();
}

//---------------------------------------------------------------------------
bool __fastcall TD2XXDriver::CharReady(void)
{
  return( IBuffUsed > 0 );
}

//---------------------------------------------------------------------------
char __fastcall TD2XXDriver::GetChar(void)
{
  char Result = 0;
  if( IBuffUsed > 0 )  GetBlock( &Result, 1 );
  return( Result );
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::PutChar( char ch)
{
  PutBlock( &ch, 1 );
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::PutString( char *s )
{
  int l = strlen(s);
  if( l > 0 )  PutBlock( s, l );
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SendBreak( WORD Ticks )
{
  if( IsEnabled() )  SetBreak(true);
  Sleep( Ticks * 55 );
  if( IsEnabled() )  SetBreak(false);
}

//---------------------------------------------------------------------------
int __fastcall TD2XXDriver::PutBlock( char *Buf, DWORD Count )
{
  if( !IsEnabled() )  return(0);

  EnterCriticalSection(&WriteSection);
  DWORD used = OBuffUsed;
  DWORD pos1 = OBuffPos;
  LeaveCriticalSection(&WriteSection);

  DWORD pos2 = pos1 + used;

  DWORD size1;
  DWORD size2 = 0;

  if( pos2 < OBuffSize ){
    size1 = OBuffSize - pos2;
    if( size1 > Count ){
      size1 = Count;
    }else{
      Count -= size1;
      size2 = pos1;
      if( size2 > Count )  size2 = Count;
    }
  }else{
    pos2 -= OBuffSize;
    size1 = pos1 - pos2;
    if( size1 > Count )  size1 = Count;
  }

  if( size1 > 0 )  CopyMemory( &OBuffer[pos2], Buf, size1 );
  if( size2 > 0 )  CopyMemory( OBuffer, &Buf[size1], size2 );

  DWORD Result = size1 + size2;

  OBuffUsed += Result;

  if( Result )  SetEvent(wtEvent);

  return( Result );
}

//---------------------------------------------------------------------------
int __fastcall TD2XXDriver::GetBlock( char *Buf, DWORD Count )
{
  if( !IsEnabled() )  return(0);

  if( IBuffUsed == 0 )  return(0);

  EnterCriticalSection(&ReadSection);
  DWORD used = IBuffUsed;
  DWORD pos1 = IBuffPos;
  LeaveCriticalSection(&ReadSection);

  DWORD pos2 = pos1 + used;

  DWORD size1;
  DWORD size2 = 0;

  if( pos2 < IBuffSize ){
    size1 = pos2 - pos1;
    if( size1 > Count )  size1 = Count;
  }else{
    pos2 -= IBuffSize;
    size1 = IBuffSize - pos1;
    if( size1 > Count ){
      size1 = Count;
    }else{
      Count -= size1;
      size2 = pos2;
      if( size2 > Count )  size2 = Count;
    }
  }

  if( size1 > 0 )  CopyMemory( Buf, &IBuffer[pos1], size1 );
  if( size2 > 0 )  CopyMemory( &Buf[size1], IBuffer, size2 );

  DWORD Result = size1 + size2;

  EnterCriticalSection(&ReadSection);
  IBuffUsed -= Result;
  IBuffPos += Result;
  if( IBuffPos >= IBuffSize )  IBuffPos -= IBuffSize;
  LeaveCriticalSection(&ReadSection);

  SetEvent(rtEvent) ;
  return( Result )  ;
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetComControl( TComControl Value )
{
  FComControl = Value;
  if( IsEnabled() ){
    FT_W32_GetCommState( FHandle, &FDCB );
    DWORD flags = DCB_BINARY;
    switch( FComControl ){
    case ccSoft: flags |= SOFT_FLOW;  break;
    case ccHard: flags |= HARD_FLOW;  break;
    }
    DWORD *tmp = (DWORD *)(&FDCB);
    tmp[2] = flags;
    FT_W32_SetCommState(FHandle, &FDCB);
  }

}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetXonLim(WORD value)
{
  FXonLim = value;
  if( IsEnabled() ){
    FT_W32_GetCommState( FHandle, &FDCB );
    FDCB.XonLim = FXonLim;
    FT_W32_SetCommState(FHandle, &FDCB);
  }
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetXoffLim(WORD value)
{
  DWORD ds = IBuffSize;
  if( ds > 65353 )  ds = 65535;
  WORD ws = (WORD)ds;
  if( value > ws )  value = ws;

  FXoffLim = value;
  if( IsEnabled() ){
    FT_W32_GetCommState( FHandle, &FDCB );
    FDCB.XoffLim = FXoffLim;
    FT_W32_SetCommState(FHandle, &FDCB);
  }

}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetXonChar(BYTE value)
{
  FXonChar = value;
  if( IsEnabled() ){
    FT_W32_GetCommState( FHandle, &FDCB );
    FDCB.XonChar = FXonChar;
    FT_W32_SetCommState(FHandle, &FDCB);
  }
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetXoffChar(BYTE value)
{
  FXoffChar = value;
  if( IsEnabled() ){
    FT_W32_GetCommState( FHandle, &FDCB );
    FDCB.XoffChar = FXoffChar;
    FT_W32_SetCommState(FHandle, &FDCB);
  }
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetEventChar(BYTE value)
{
  FEventChar = value;
  if( IsEnabled() ){
    FT_W32_GetCommState( FHandle, &FDCB );
    FDCB.EvtChar = FEventChar;
    FT_W32_SetCommState(FHandle, &FDCB);
  }
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetBaudRate( TBaudRate Value )
{
  int CBR[15] = {CBR_110,
                 CBR_300,
                 CBR_600,
                 CBR_1200,
                 CBR_2400,
                 CBR_4800,
                 CBR_9600,
                 CBR_14400,
                 CBR_19200,
                 CBR_38400,
                 CBR_56000,
                 CBR_57600,
                 CBR_115200,
                 CBR_128000,
                 CBR_256000 };

  FBaudRate = Value;
  if( IsEnabled() ){
    FT_W32_GetCommState( FHandle, &FDCB );
    FDCB.BaudRate = CBR[FBaudRate];
    FT_W32_SetCommState(FHandle, &FDCB);
  }
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetParity( TParity Value )
{
  char PAR[5] = {NOPARITY,
                 ODDPARITY,
                 EVENPARITY,
                 MARKPARITY,
                 SPACEPARITY };
  FParity = Value;
  if( IsEnabled() ){
    FT_W32_GetCommState( FHandle, &FDCB );
    FDCB.Parity = PAR[FParity] ;
    FT_W32_SetCommState( FHandle, &FDCB );
  }
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetStopBits( TStopBits Value )
{
  char STB[3] = {ONESTOPBIT, ONE5STOPBITS, TWOSTOPBITS};

  FStopBits = Value;
  if( IsEnabled() ){
    FT_W32_GetCommState(FHandle, &FDCB);
    FDCB.StopBits = STB[FStopBits];
    FT_W32_SetCommState(FHandle, &FDCB);
  }

  BitsWarning();
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetDataBits( TDataBits Value )
{
  if( Value < 5 || Value > 8 )  Value = 8;

  FDataBits = Value;
  if( IsEnabled() ){
    FT_W32_GetCommState(FHandle, &FDCB);
    FDCB.ByteSize = Value;
    FT_W32_SetCommState(FHandle, &FDCB);
  }

  BitsWarning();
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::BitsWarning(void)
{
  if( FDataBits != 5 ){
    if( FStopBits == sb1_5 )
      MessageBox(NULL ,"Invalid combination (6,7 or 8 data bits with 1.5 stop bits)", " USB ",  MB_OK | MB_ICONWARNING );
  }else{
    if( FStopBits == sb2_0 )
      MessageBox(NULL ,"Invalid combination (5 data bits with 2 stop bits)", " USB ",  MB_OK | MB_ICONWARNING );
  }
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetBaud( int B )
{
  switch( B ){
    case 110     : BaudRate = cbr110;    break;
    case 300     : BaudRate = cbr300;    break;
    case 600     : BaudRate = cbr600;    break;
    case 1200    : BaudRate = cbr1200;   break;
    case 2400    : BaudRate = cbr2400;   break;
    case 4800    : BaudRate = cbr4800;   break;
    case 9600    : BaudRate = cbr9600;   break;
    case 14400   : BaudRate = cbr14400;  break;
    case 19200   : BaudRate = cbr19200;  break;
    case 38400   : BaudRate = cbr38400;  break;
    case 56000   : BaudRate = cbr56000;  break;
    case 57600   : BaudRate = cbr57600;  break;
    case 115200  : BaudRate = cbr115200; break;
    case 128000  : BaudRate = cbr128000; break;
    case 256000  : BaudRate = cbr256000; break;
  }
}

//---------------------------------------------------------------------------
int __fastcall TD2XXDriver::GetBaud(void)
{
  int result = 0;
  switch( BaudRate ){
    case cbr110     : result = 110;    break;
    case cbr300     : result = 300;    break;
    case cbr600     : result = 600;    break;
    case cbr1200    : result = 1200;   break;
    case cbr2400    : result = 2400;   break;
    case cbr4800    : result = 4800;   break;
    case cbr9600    : result = 9600;   break;
    case cbr14400   : result = 14400;  break;
    case cbr19200   : result = 19200;  break;
    case cbr38400   : result = 38400;  break;
    case cbr56000   : result = 56000;  break;
    case cbr57600   : result = 57600;  break;
    case cbr115200  : result = 115200; break;
    case cbr128000  : result = 128000; break;
    case cbr256000  : result = 256000; break;
  }
  return( result );
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::FlushInBuffer(void)
{
  if( IsEnabled() ){
    FT_W32_PurgeComm( FHandle, PURGE_RXABORT | PURGE_RXCLEAR );
  }
  EnterCriticalSection(&ReadSection);
  IBuffUsed = 0;
  LeaveCriticalSection(&ReadSection);
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::FlushOutBuffer(void)
{
  if( IsEnabled() ){
    FT_W32_PurgeComm( FHandle, PURGE_TXABORT | PURGE_TXCLEAR );
  }
  EnterCriticalSection(&WriteSection);
  ResetEvent(wtEvent);
  OBuffUsed = 0;
  LeaveCriticalSection(&WriteSection);
}

//---------------------------------------------------------------------------
bool __fastcall TD2XXDriver::GetModemState( int Index )
{
  DWORD Flag, State;

  switch( Index ){
    case 1:  State = MS_CTS_ON;  break;
    case 2:  State = MS_DSR_ON;  break;
    case 3:  State = MS_RING_ON; break;
    case 4:  State = MS_RLSD_ON; break;
    default: State = 0;          break;
  }

  bool Result = false;
  if( IsEnabled() ){
    if( FT_W32_GetCommModemStatus(FHandle, &Flag) )
      Result = (Flag & State);
  }
  return( Result );
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::PowerOn(void)
{
  FT_W32_EscapeCommFunction( FHandle, SETDTR );
  FT_W32_EscapeCommFunction( FHandle, SETRTS );
  FRTS = true;
  FDTR = true;
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::PowerOff(void)
{
  FT_W32_EscapeCommFunction( FHandle, CLRDTR );
  FT_W32_EscapeCommFunction( FHandle, CLRRTS );
  FRTS = false;
  FDTR = false;
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetDTR( bool State )
{
  if( FOpen ){
    if( FT_W32_EscapeCommFunction( FHandle, State ? SETDTR : CLRDTR ) ){
      FDTR = State;
    }
  }else{
    FDTR = State;
  }
}

//---------------------------------------------------------------------------
void __fastcall TD2XXDriver::SetRTS( bool State )
{
  if( FOpen ){
    if( FT_W32_EscapeCommFunction( FHandle, State ? SETRTS : CLRRTS ) ){
      FRTS = State;
    }
  }else{
    FRTS = State;
  }
}

//---------------------------------------------------------------------------
bool __fastcall TD2XXDriver::SetBreak( bool State )
{
  bool Result = false;
  if( FOpen )
    Result = FT_W32_EscapeCommFunction( FHandle, State ? SETBREAK : CLRBREAK );

  return( Result );
}
//---------------------------------------------------------------------

void __fastcall TD2XXDriver::GetDevicesList(TStringList* DeviceList)
{
  DeviceList->Clear() ;

  //    
  FT_STATUS ftStatus;
  DWORD numDevs;
  ftStatus = FT_ListDevices(&numDevs,NULL,FT_LIST_NUMBER_ONLY);
  if (ftStatus == FT_OK)
    {
      if ( !numDevs ) return ;
    }
  else
    {
     throw "   FT " ;
    }
  //     
  DWORD*  BufsList = new  DWORD  [numDevs] ;
  for (unsigned int DevIndex = 0 ; DevIndex < numDevs ; DevIndex++)
   {
     BufsList[DevIndex] = (DWORD) new char [1024] ;
   }

  ftStatus = FT_ListDevices(BufsList,&numDevs,FT_LIST_ALL|FT_OPEN_BY_DESCRIPTION);
  if (ftStatus == FT_OK)
     {
       for (unsigned int DevIndex = 0 ; DevIndex < numDevs ; DevIndex++)
        {
          DeviceList -> Add ((char*)BufsList[DevIndex])  ;
        }
     }
  else
     {
       throw "    FT " ;
     }

  //     
  
  for ( unsigned int DevIndex = 0 ; DevIndex < numDevs ; DevIndex++)
   {
     delete ((char*)BufsList[DevIndex])  ;
   }
  delete BufsList ;
}
